import {arrayFindIndex} from "./utils/util";
import {getCell, getColumnByCell, getRowIdentity} from "./utils/util";
import {getStyle, hasClass, removeClass, addClass} from "./utils/dom";
import {debounce} from "throttle-debounce";
import LayoutObserver from "./layout-observer";
import {mapStates} from "./store/helper";
import TableCell from "./table-cell.vue";
import TableCellOperation from "./table-cell-operation.vue";
import TableRow from "./table-row.js";

export default {
  name: "TableBody",
  render(h) {
    const data = this.data || [];
    return (
      <tbody class="xt-table__body">
      {
        data.reduce((acc, row) => acc.concat(this.wrappedRowRender(row, acc.length)), [])
      }
      </tbody>
    );
  },
  methods: {

    /*#region 自定义函数(组件change事件等)*/
    validate(callback) {
      const validatorStates = [];
      this.$refs && Object.keys(this.$refs).map((keyItem) => {
        if (this.selectFlag) {
          if (this.$refs[keyItem] && this.$refs[keyItem].isSelected) {
            this.$refs[keyItem] && this.$refs[keyItem].validate && this.$refs[keyItem].validate((validatorState) => {
              validatorStates.push(validatorState);
            });
          }
        } else {
          this.$refs[keyItem] && this.$refs[keyItem].validate && this.$refs[keyItem].validate((validatorState) => {
            validatorStates.push(validatorState);
          });
        }
      });
      const returnValidate = validatorStates.includes(false);
      if (returnValidate) {callback(false);} else {callback(true);}
    },
    resetFields(callback) {
      this.$refs && Object.keys(this.$refs).map((keyItem) => {
        this.$refs[keyItem] && this.$refs[keyItem].resetFields && this.$refs[keyItem].resetFields((validatorState) => {});
      });
      callback();
    },
    getKeyOfRow(row, index) {
      const rowKey = this.table.rowKey;
      if (rowKey) {
        return getRowIdentity(row, rowKey);
      }
      return index;
    },

    /*#endregion*/

    /*#region 操作函数(handle)*/

    /*#endregion*/

    rowRender(row, $index, treeRowData) {
      const rowKey = this.getKeyOfRow(row, $index);
      return (
        <table-row
          ref={"tableRow__" + rowKey}
          bodyThis={this}
          row={row}
          $index={$index}
          treeRowData={treeRowData}
          table={this.table}
          store={this.store}
          key={ rowKey }
        ></table-row>
      );
    },
    wrappedRowRender(row, $index) {
      const store = this.store;
      const { isRowExpanded, assertRowKey } = store;
      const { treeData, lazyTreeNodeMap, childrenColumnName, rowKey } = store.states;
      if (this.hasExpandColumn && isRowExpanded(row)) {
        const renderExpanded = this.table.renderExpanded;
        const tr = this.rowRender(row, $index);
        if (!renderExpanded) {
          console.error("[Element Error]renderExpanded is required.");
          return tr;
        }
        // 使用二维数组，避免修改 $index
        return [[
          tr,
          <tr key={"expanded-row__" + tr.key}>
            <td colspan={ this.columnsCount } class="xt-table__expanded-cell">
              { renderExpanded(this.$createElement, { row, $index, store: this.store }) }
            </td>
          </tr>]];
      } else if (Object.keys(treeData).length) {
        assertRowKey();
        // TreeTable 时，rowKey 必须由用户设定，不使用 getKeyOfRow 计算
        // 在调用 rowRender 函数时，仍然会计算 rowKey，不太好的操作
        const key = getRowIdentity(row, rowKey);
        let cur = treeData[key];
        let treeRowData = null;
        if (cur) {
          treeRowData = {
            expanded: cur.expanded,
            level: cur.level,
            display: true
          };
          if (typeof cur.loaded === "boolean" && cur.loaded) {
            treeRowData.noLazyChildren = !(cur.children && cur.children.length);
          }
          if (typeof cur.loading === "boolean") {
            treeRowData.loading = cur.loading;
          }
        }
        const tmp = [this.rowRender(row, $index, treeRowData)];
        // 渲染嵌套数据
        if (cur) {
          // currentRow 记录的是 index，所以还需主动增加 TreeTable 的 index
          let i = 0;
          const traverse = (children, parent) => {
            if (!(children && children.length && parent)) {return;}
            children.forEach((node) => {
              // 父节点的 display 状态影响子节点的显示状态
              const innerTreeRowData = {
                display: parent.display && parent.expanded,
                level: parent.level + 1
              };
              const childKey = getRowIdentity(node, rowKey);
              if (childKey === undefined || childKey === null) {
                throw new Error("for nested data item, row-key is required.");
              }
              cur = { ...treeData[childKey] };
              // 对于当前节点，分成有无子节点两种情况。
              // 如果包含子节点的，设置 expanded 属性。
              // 对于它子节点的 display 属性由它本身的 expanded 与 display 共同决定。
              if (cur) {
                innerTreeRowData.expanded = cur.expanded;
                // 懒加载的某些节点，level 未知
                cur.level = cur.level || innerTreeRowData.level;
                cur.display = !!(cur.expanded && innerTreeRowData.display);
                if (typeof cur.loaded === "boolean" && cur.loaded) {
                  innerTreeRowData.noLazyChildren = !(cur.children && cur.children.length);
                }
                if (typeof cur.lazy === "boolean") {
                  innerTreeRowData.loading = cur.loading;
                }
              }
              i++;
              tmp.push(this.rowRender(node, $index + i, innerTreeRowData));
              if (cur) {
                const nodes = lazyTreeNodeMap[childKey] || node[childrenColumnName];
                traverse(nodes, cur);
              }
            });
          };
          // 对于 root 节点，display 一定为 true
          cur.display = true;
          const nodes = lazyTreeNodeMap[key] || row[childrenColumnName];
          traverse(nodes, cur);
        }
        return tmp;
      } else {
        return this.rowRender(row, $index);
      }
    }
  },
  components: {
    TableCell,
    TableCellOperation,
    TableRow
  },
  computed: {
    table() {
      return this.$parent;
    },

    ...mapStates({
      data: "data",
      columns: "columns",
      treeIndent: "indent",
      leftFixedLeafCount: "fixedLeafColumnsLength",
      rightFixedLeafCount: "rightFixedLeafColumnsLength",
      columnsCount: (states) => states.columns.length,
      leftFixedCount: (states) => states.fixedColumns.length,
      rightFixedCount: (states) => states.rightFixedColumns.length,
      hasExpandColumn: (states) => states.columns.some(({ type }) => type === "expand"),
      selectFlag: "selectFlag"
    }),

    firstDefaultColumnIndex() {
      return arrayFindIndex(this.columns, ({ type }) => type === "default");
    }
  },
  data() {
    return {
      tooltipContent: ""
    };
  },
  mixins: [LayoutObserver],
  watch: {
    "store.states.hoverRow"(newVal, oldVal) {
      if (!this.store.states.isComplex || this.$isServer) {return;}
      let raf = window.requestAnimationFrame;
      if (!raf) {
        raf = (fn) => setTimeout(fn, 16);
      }
      raf(() => {
        const rows = this.$el.querySelectorAll(".xt-table__row");
        const oldRow = rows[oldVal];
        const newRow = rows[newVal];
        if (oldRow) {
          removeClass(oldRow, "hover-row");
        }
        if (newRow) {
          addClass(newRow, "hover-row");
        }
      });
    }
  },
  props: {
    store: {
      required: true
    },
    stripe: Boolean,
    context: {},
    rowClassName: [String, Function],
    rowStyle: [Object, Function],
    fixed: String,
    highlight: Boolean
  }
};
